#ifndef VEC3D_HH
#define VEC3D_HH

/*
  James R. Diebel
  Stanford University
  
  Started: 23 August 2004
  Last revised: 27 Sugust 2004

  vec3d.hh - Class definition of basic 3d vector class

  Depends on nothing

  The only point of this class is to allow for convenient vector notation
  and to thereby make the code easier to read

  -- Vectors are initialized with "Vec3d(x,y,z)" or "Vec3d()", which yields 
     the zero vector.
  -- The values of existing vectors may be set with "v[0] = x" or
     "v.set(x,y,z)" or "v1 = v2" or "v = d" where d is a float[3]
  -- Values may be referenced with the [] notation too: "x = v[0]"
  -- Vectors add and subtract like scalars: "u = v + w; u -= v;"
  -- Most reasonable operations with scalars are allowed: "v = a*u + b" or 
     "w = (-v + a)/b" or  "v = (-(a - u)/b)+c"
  -- Dot and cross products are defined: "u.dot(v)" and "u.cross(v)"
  -- Equality of two vectors may be tested with "if (u==v || u!=v)" as usual
  -- Relative vectors may be formed with "u.to(v)" or "v.from(u)" which are
     both equal to "(v-u)" and are convenient when thinking of vectors as
     arrows between points in space: the arrow from u to v is "u.to(v)"
  -- The length of a vector can be computed as "v.len()" or "~v"
  -- The distance from one vector to another can be computed as
     "v.distTo(u)" which is the same as "(v-u).len()" or "~(v-u)"
  -- The squared length or squared distance can be computed with the
     related "v.len2()" and "v.dist2To(u)"
  -- A vector may be normalized with "v.normalize()"
  -- A vector may be printed with a standard "cout << v << endl"
     or "v.print(); cout << endl" or "v.printcr()", and to print the length
     too, "v.printall()"
*/

#include <math.h>
#include <stdlib.h>
#include <iostream>
#ifndef PI
#define PI 3.14159265358979323846
#endif

namespace bmtk {

using namespace std;

class Vec3d {
 public: // everything is public

  // The data is a simple array of size 3
  float x[3];
  
  // Constructors
  Vec3d();
  Vec3d(const float x0, const float x1, const float x2);
  Vec3d(const float a);
  Vec3d(const Vec3d& v);
  Vec3d(const float* x_);

  // Assignments
  void set(const float x0, const float x1, const float x2);
  Vec3d operator = (const Vec3d &v);
  Vec3d operator = (const float d[3]);
  Vec3d operator = (float a);
  Vec3d operator = (int a);

  // Reference operator: zero-based square bracket referencing
  float operator [] (const int index) const;
  float& operator [] (const int index);

  // Return relative vectors (vector subtraction)
  Vec3d to(const Vec3d &v) const;
  Vec3d from(const Vec3d &v) const;

  // Addition and subtraction: <vector> +/- <vector or scalar>
  // note: does not perform <scalar> +/- <vector>
  //       (see non-class functions below)
  Vec3d operator + (const Vec3d &v) const;
  Vec3d operator + (const float a) const;
  void operator += (const Vec3d &v);
  void operator += (const float a);
  Vec3d operator - (const Vec3d &v) const;
  Vec3d operator - (const float a) const;
  void operator -= (const Vec3d &v);
  void operator -= (const float a);
  Vec3d operator - () const;
    
  // Multiplication and division by scalars: <vector> *// <scalar>
  // note: does not perform <scalar> * <vector>
  //       (see non-class functions below)
  Vec3d operator * (const float a) const;
  Vec3d operator / (const float a) const;
  Vec3d operator * (const Vec3d v) const;
  Vec3d operator / (const Vec3d v) const;
  Vec3d operator *= (const float a);
  Vec3d operator /= (const float a);
  Vec3d operator *= (const Vec3d v);
  Vec3d operator /= (const Vec3d v);

  // Dot and Cross Products
  float dot(const Vec3d &v) const;
  Vec3d cross(const Vec3d &v) const;

  // Equality tests make sense
  bool operator == (const Vec3d &v) const;
  bool operator != (const Vec3d &v) const;
  
  // Length and Normalize
  float len(void) const;
  float operator ~ (void) const;
  float len2(void) const;
  float distTo(Vec3d &v) const;
  float dist2To(Vec3d &v) const;
  void normalize(void);
  
  // Print routines
  void print(void);
  void printcr(void);
  void printall(void);
};

// Define a few global operators to handle cases where 
// a Vec3d is the second argument in a binary operator
Vec3d operator + (const float a, const Vec3d &v);
Vec3d operator - (const float a, const Vec3d &v);
Vec3d operator * (const float a, const Vec3d &v);

// Define << operator to handle output to screen
ostream& operator << (ostream& os, const Vec3d &v);

} // namespace bmtk

#endif
