#ifndef VERT_HH
#define VERT_HH

/*
  James R. Diebel
  Stanford University
  
  Started: 27 August 2004

  vert.hh - class definition for a vertex in a 3d surface mesh as defined
  in the header file mesh.hh.

  Depends on:
  - Vec3d class (vec3d.hh)
  - Edge class (edge.hh)
  - Face class (face.hh)

  Used almost exclusively by Mesh class (mesh.hh)
*/

#include <iostream>
#include "vec3d.hh"
#include "mat3x3.hh"
#include "mesh.hh"
#include "edge.hh"
#include "face.hh"

namespace bmtk {

using namespace std;

class Vert {
 public:
  // Data
  int nf,ne; // number of adjacent faces, edges
  int i; // index of self within global vertex list
  float l; // average length of connected edges
  bool bound; // whether the vertex is on the mesh boundary
  int flag; // used to mark vertices for region boundary tracing
  int marker; // the same by a different name
  float depth; // depth from some specified origin
  float psi; // source potential to original vertex position
  float psiEdge; // sum of edge potential in vertex neighborhood
  float psiLocal; // psi + psiEdge
  float psiLocal0; // old value
  Vec3d gradPsi; // gradient of psi WRT x
  Vec3d gradPsi0; // old value of gradient
  float normGradPsi; // 2-norm of the gradient
  float normGradPsi0; // old value of the 2-norm of the gradient
  float step; // local step size
  Vec3d dir; // search direction for CG steps
  Vec3d dir0; // search direction of previous iteration
  Vec3d x0; // original position of vertex in 3-space
  Vec3d x; // position of vertex in 3-space
  Vec3d xr; // reference vector used in CG optimization
  Vec3d* xs; // array of the (m->nd) different positions
  Vec3d n; // normal vector assigned to vertex
  Vec3d c; // color in RGB-space
  Mat3x3 iCov; // inverse variance/covariance matrix

  class Mesh *m; // Link to parent mesh
  class Face **f; // list of faces that include this vertex
  class Edge **e; // list of edges that include this vertex
  class Vert **v; // list of vertices that share an edge with this vertex

  // Constructor/Destructor
  Vert();
  ~Vert();

  // Member functions
  void findNormal();
  void findLength();
  void findPotential();
  void findDepth();
  void refreshLocale();
  void findLocalEdgePotential();
  void findGradient();
  void findSearchDir(float beta);
  void move(float globalStep);
  void move();
  void flatten();

  // Equality tests
  inline bool operator == (const Vert &vp) const;
  inline bool operator != (const Vert &vp) const;
};

// Define << operator to handle output to screen
ostream& operator << (ostream& os, const Vert &v);

} // namespace bmtk

#endif
